from datetime import timedelta
import math
import re
from typing import List

import torch
from torch.distributions.gamma import Gamma
from torch.distributions.categorical import Categorical
from kismet.personality.responses import responses as responses_
from kismet.types import Message


class Responder:
    def __init__(self, responses=responses_):
        self.categorical = Categorical(torch.ones(len(responses_)))
        self.responses = responses

    def respond(self, excitement: int):
        if excitement < 1:
            return None

        response = [
            self.responses[self.categorical.sample()]
            for _ in range(excitement)
        ]
        return " ".join(response)


responder_ = Responder()

KISMET_TIMEDELTA = timedelta(seconds=32)
KISMET_PATTERN = re.compile(r"[Kk]+\s*[Ii]+\s*[Ss]+\s*[Mm]+\s*[Ee]+\s*[Tt]+")

def analyze(messages: List[Message], client_id: int):
    if len(messages) == 0:
        return None
    count = len(messages)
    latest = messages[0].created_at
    messages.reverse()
    mentioned = 0
    attention = 0
    for idx, message in enumerate(messages):
        is_result = (message.author_id == client_id and message.reply_id)
        if is_result:
            mentioned = idx
        elif is_mentioned(message.content):
            mentioned = idx
        delta = latest - message.created_at
        attention += (
            get_attention(message.content)
            * (0.25 if is_result else 1)
            * 4/(4+delta.seconds)
            * (2/(2*(1 + idx - mentioned)))
            * (2/(2*(1 + count - idx)))
        )
    excitement = round(
        float(Gamma(
            1 + (2/(1+math.exp(2 * (count - mentioned - 1)))),
            1 / (2 * attention)
        ).sample())
        if attention > 0 else 0
    )
    return respond(excitement)

def respond(excitement: int, responder: Responder = responder_):
    return responder.respond(excitement)

def is_mentioned(string: str):
    return KISMET_PATTERN.match(string)

def get_attention(string: str):
    return 1 / (1 + math.exp(-len(string.replace(r"\s", ""))/4))


from dataclasses import dataclass
from datetime import datetime
from typing import Optional

@dataclass
class Message:
  message_id: int
  author_id: int
  created_at: datetime
  edited_at: Optional[datetime]
  reply_id: Optional[int]
  content: str


from os import getenv
from discord import Client, Message as DiscordMessage

from kismet.core import process_markdown, process_messages
from kismet.types import Message

token = getenv("DISCORD_TOKEN", "")
clientid = getenv("DISCORD_CLIENTID", "0")
permissions = getenv("DISCORD_PERMISSIONS", "377957238848")

oauth2_template = (
    "https://discordapp.com/oauth2/authorize?scope=bot&client_id=%s&permissions=%s"
)
oauth2_url = oauth2_template % (clientid, permissions)
print("Use the following URL to invite:")
print(oauth2_url)


# Define client
client = Client()

def replace_mentions(string: str):
    return string.replace("<@" + str(client.user.id) + ">", "Kismet")

def convert_message(message: DiscordMessage):
    return Message(
        message_id=message.id,
        author_id=message.author.id,
        created_at=message.created_at,
        edited_at=message.edited_at,
        reply_id=message.reference.message_id if message.reference else None,
        content=message.content,
    )

@client.event
async def on_message(event):
    if event.author == client.user:
        return
    else:
        response = process_markdown(event.content)
        if response:
            await event.reply(response)
        channel = event.channel
        history = [convert_message(message) async for message in channel.history(limit=16)]
        reply = process_messages(history, client.user.id)
        if reply:
            await event.channel.send(reply)


@client.event
async def on_ready():
    print("Logged in as")
    print(client.user.name)
    print(client.user.id)
    print("------")


client.run(token)
